package net.fukure.gae.common;

import static com.google.appengine.api.datastore.FetchOptions.Builder.withLimit;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.google.appengine.api.datastore.DatastoreFailureException;
import com.google.appengine.api.datastore.DatastoreService;
import com.google.appengine.api.datastore.DatastoreServiceFactory;
import com.google.appengine.api.datastore.DatastoreTimeoutException;
import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.EntityNotFoundException;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.appengine.api.datastore.Query;

public class EntityUtil {
	
	public static Entity getEntityOfKind(String kind,int offset) throws IOException, EntityNotFoundException{
		Entity taskEntity = null;
		Query query = new Query(kind);
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (Iterator<Entity> iterator = datastore.prepare(query).asList(withLimit(1).offset(offset)).iterator(); iterator.hasNext();) {
			taskEntity = (Entity) iterator.next();
		}
		if(taskEntity==null){
			throw new EntityNotFoundException(null);
		}

		return taskEntity;
	}
	
	public static List<Entity> getAllEntityOfKind(String kind) throws IOException{
		return getAllEntityOfKind(kind, 0, 1000);
	}
	
	public static List<Entity> getAllEntityOfKind(String kind, int offset, int limit) {
		Query query = new Query(kind).addSort("__key__",Query.SortDirection.DESCENDING);
		List<Entity> list = new ArrayList<Entity>();
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (Iterator<Entity> iterator = datastore.prepare(query).asList(withLimit(limit).offset(offset)).iterator(); iterator.hasNext();) {
			Entity taskEntity = (Entity) iterator.next();
			list.add(taskEntity);
		}
		return list;
	}

	public void deleteAllEntityOfKind(String kind){
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		Query query = new Query(kind);
		for (Entity taskEntity : datastore.prepare(query).asIterable()) {
			datastore.delete(taskEntity.getKey());
		}
	}
	
	public static byte[] entitySerialize(Entity taskEntity) throws IOException{
		ByteArrayOutputStream outFile = new ByteArrayOutputStream();
		ObjectOutputStream outObject = new ObjectOutputStream(outFile);
		outObject.writeObject(taskEntity);
		return outFile.toByteArray();
	}
	
	public static Entity entityDeserialize(byte[] byteArray) throws IOException, ClassNotFoundException{
		ByteArrayInputStream inFile = new ByteArrayInputStream(byteArray);
		ObjectInputStream inObject = new ObjectInputStream(inFile);
		return (Entity) inObject.readObject();
	}
	
	public static Entity getEntity(String keyString) throws IOException, EntityNotFoundException{
		Key key = KeyFactory.stringToKey(keyString);
		return getEntity(key);
	}
	
	public static Entity getEntity(Key key) throws IOException, EntityNotFoundException{
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (int i = 0; i < 5; i++) {
			try {
				return datastore.get(key);
			} catch (DatastoreTimeoutException e) {
				continue;
			} catch (DatastoreFailureException e) {
				continue;
			}
		}
		throw new IOException();
	}
	
	public static Entity getEntity(String kind, long id) throws IOException, EntityNotFoundException{
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (int i = 0; i < 5; i++) {
			try {
				Key key = KeyFactory.createKey(kind,id);
				return datastore.get(key);
			} catch (DatastoreTimeoutException e) {
				continue;
			} catch (DatastoreFailureException e) {
				continue;
			}
		}
		throw new IOException();
	}

	public static void putEntity(Entity entity) throws IOException, ClassNotFoundException {
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (int i = 0; i < 5; i++) {
			try {
				datastore.put(entity);
				return;
			} catch (DatastoreTimeoutException e) {
				continue;
			} catch (DatastoreFailureException e) {
				continue;
			}
		}
		throw new IOException();
	}
	
	public static void putEntity(byte[] byteArray) throws IOException, ClassNotFoundException {
		Entity entity = entityDeserialize(byteArray);
		putEntity(entity);
	}
	
	public static void removeEntity(Entity entity) throws IOException{
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		for (int i = 0; i < 5; i++) {
			try {
				datastore.delete(entity.getKey());
				return;
			} catch (DatastoreTimeoutException e) {
				continue;
			} catch (DatastoreFailureException e) {
				continue;
			}
		}
		throw new IOException();
	}
}
