package net.fukure.gae.csvloader;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.lang.reflect.Method;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import net.fukure.gae.common.PMF;
import net.fukure.gae.common.TestEnvironment;

import com.google.appengine.api.datastore.Blob;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.Link;
import com.google.appengine.api.datastore.ShortBlob;
import com.google.appengine.api.datastore.Text;
import com.google.appengine.api.datastore.KeyFactory.Builder;
import com.google.appengine.api.datastore.dev.LocalDatastoreService;
import com.google.appengine.api.users.User;
import com.google.appengine.repackaged.com.google.common.util.Base64;
import com.google.appengine.tools.development.ApiProxyLocalImpl;
import com.google.apphosting.api.ApiProxy;

import au.com.bytecode.opencsv.*;

public class CSVtoJDOTest {

	//CSV file name
	static private String csvFile = "TestData.txt";
	static private String blob_type = "mysql_hex";
	static private String DATASTORE_V3 = "datastore_v3";
	
	public static void main(String[] args) throws Exception {

		CSVtoJDOTest jdo = new CSVtoJDOTest();
		jdo.csvtest();

	}

	private Object setJDOParam(Class<?> gaeJDOClass,HashMap<String,String> map)
		throws CSVtoJDOClassConvertException, CSVtoJDOSetMethodException, Exception {
		
		Object obj = null;
		obj = gaeJDOClass.newInstance();
		
		Iterator<String> it = map.keySet().iterator();
		while (it.hasNext()) {
			String setter = (String) it.next();
			String value = map.get(setter);
			setJDOParamValue(obj, setter, value);
		}

		return obj;
		
	}
	
	private void setJDOParamValue(Object obj, String setter, String value)
		throws CSVtoJDOClassConvertException, CSVtoJDOSetMethodException {
	
		HashMap<String, Method> map = getMethodMap(obj);
		if(map.containsKey(setter)){
			Method setMethod = map.get(setter);
			Class<?>[] types = setMethod.getParameterTypes();
			if(types.length>0){
				Object convValue = getClassConvertedValue(types[0],value);
				try{
					setMethod.invoke(obj, convValue);
				}catch(Exception e){
					throw new CSVtoJDOSetMethodException();
				}
			}
		}
	}

	private Object getClassConvertedValue(Class<?> toClass,String value) throws CSVtoJDOClassConvertException{
		
		try{
			//System.out.println(value.getClass().toString()+"->"+toClass.toString());
			if(toClass==java.lang.String.class){
				return value;
			}
			if(toClass==com.google.appengine.api.datastore.ShortBlob.class){
				byte[] data;
				if(blob_type.equals("mysql_hex")){
					data = hexStringToByteArray(value);
				}else{
					data = Base64.decode(value);
				}
				return new ShortBlob(data);
			}
			if(toClass==java.lang.Boolean.class){
				return Boolean.valueOf(value);
			}
			if(toClass==java.lang.Short.class ||
					toClass==java.lang.Integer.class||
					toClass==java.lang.Long.class){
				return Long.parseLong(value);
			}
			if(toClass==java.lang.Float.class ||
					toClass==java.lang.Double.class)
			{
				return Double.parseDouble(value);
			}
			if(toClass==com.google.appengine.api.users.User.class){
				return new User(value,"gmail.com");
			}
			if(toClass==com.google.appengine.api.datastore.Text.class){
				return new Text(value);
			}
			if(toClass==com.google.appengine.api.datastore.Blob.class){
				byte[] data;
				if(blob_type.equals("mysql_hex")){
					data = hexStringToByteArray(value);
				}else{
					data = Base64.decode(value);
				}
				return new Blob(data);
			}
			if(toClass==java.util.Date.class){
				SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
				return (Date)sdf.parse(value);
			}
			if(toClass==com.google.appengine.api.datastore.Link.class){
				return new Link(value);
			}
			if(toClass==com.google.appengine.api.datastore.Key.class){
				return keyPathStringToKey(value);
			}
			return null;
		}catch(Exception e){
			throw new CSVtoJDOClassConvertException();
		}
	}

	public byte[] hexStringToByteArray(String hexString){
		if(hexString==null) return null;
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		int st = 0;
		if(hexString.length()>=2 && hexString.substring(0, 2).equals("0x")){
			st = 1;
		}
		for (int i = st; i < hexString.length()/2; i++) {
			int b = Integer.parseInt(hexString.substring(i*2, i*2+2), 16);
			baos.write(b);
		}
		return baos.toByteArray();
	}
	
	private Class<?> getClass(String className) throws ClassNotFoundException{
		if(className.indexOf(".")>0){
			return Class.forName(className);
		}
		return Class.forName(this.getClass().getPackage().getName()+"."+className);
	}

	private HashMap<String, Method> getMethodMap(Object obj){
		HashMap<String, Method> map = new HashMap<String, Method>();
		
		Method[] methods = obj.getClass().getMethods();
		
		for (int i = 0; i < methods.length; i++) {
			Method method = methods[i];
			map.put(method.getName(), method);
		}
		
		return map;
	}

	Key getParentKey(String[] parentPath, String[] csvline){
		Builder builder = null;
		try{
			builder = new Builder(parentPath[0], Long.parseLong(csvline[0]));
		}catch(Exception e){
			builder = new Builder(parentPath[0], csvline[0]);
		}
		for (int i = 1; i < parentPath.length; i++) {
			try{
				builder.addChild(parentPath[i], Long.parseLong(csvline[i]));
			}catch(Exception e){
				builder.addChild(parentPath[i], csvline[i]);
			}
		}
		return builder.getKey();
	}

	Key keyPathStringToKey(String pathString){
		String[] path = pathString.split("/");
		Builder builder = null;
		for (int i = 0; i < path.length; i++) {
			Pattern pattern = Pattern.compile("([^\\(]+)\\(\"?([^\\)]+\"?)\\)");
			Matcher matcher = pattern.matcher(path[i]);
			String kind = null;
			String id = null;
			while (matcher.find()) {
				kind = matcher.group(1);
				id = matcher.group(2);
				break;
			}
			if(builder==null){
				try{
					//Ex. MyData(25)
					long ld = Long.parseLong(id);
					builder = new Builder(kind,ld);
				}catch(Exception e){
					String[] parts = id.split("\"");
					if(parts.length>1){
						//Ex. MyData("25")
						builder = new Builder(kind,parts[1]);
					}else{
						//Ex. MyData("abc")
						builder = new Builder(kind,parts[0]);
					}
				}
			}else{
				try{
					long ld = Long.parseLong(id);
					builder.addChild(kind,ld);
				}catch(Exception e){
					builder.addChild(kind,id);
				}
			}
		}
		if(builder==null){
			return null;
		}else{
			return builder.getKey();
		}
	}
	
	private class CSVtoJDOParser {
		ArrayList<String> keyPath = new ArrayList<String>();
		ArrayList<String> params = new ArrayList<String>();
		String parentClassName = null;
		String parentClassMethodName = null;
		int propIndex = 0;
		
		String[] getParentPath(){
			return keyPath.toArray(new String[0]);
		}
		
		void parseHeaderLine(String[] setterLine){
			for (int i = 0; i < setterLine.length; i++) {
				String str = setterLine[i];
				if(str.lastIndexOf("()")>0){
					int end = str.lastIndexOf("()");
					str = str.substring(0, end);
					//親クラスはクラスとメソッドが記述されているので分ける
					if(str.lastIndexOf(".")>0){
						String[] part = str.split("\\.");
						parentClassName = part[part.length-2];
						parentClassName = parentClassName.trim();
						parentClassMethodName = part[part.length-1];
						parentClassMethodName = parentClassMethodName.trim();
						keyPath.add(parentClassName);
						propIndex = i + 1;
					}else{
						//メソッドのみ パラメータに追加
						params.add(str.trim());
					}
				}else{
					//クラスのみ キーパスに追加
					 keyPath.add(str.trim());
				}
			}
		}
		
		HashMap<String, String> getParamsMap(String[] line){
			HashMap<String,String> list = new HashMap<String,String>();
			for (int i = 0; i < params.size(); i++) {
				list.put(params.get(i),line[propIndex+i].trim());
			}
			return list;
		}
	}

	private void startUP(){
		TestEnvironment environment = new TestEnvironment();
		
		ApiProxyLocalImpl localproxy = new ApiProxyLocalImpl(new File("war")) {};
		ApiProxy.setEnvironmentForCurrentThread(environment);

		ApiProxy.setDelegate(localproxy);
	}
	
	private void downEnd(){
		try{
			ApiProxyLocalImpl localproxy = (ApiProxyLocalImpl) ApiProxy.getDelegate();
			LocalDatastoreService localstore = (LocalDatastoreService)localproxy.getService(DATASTORE_V3);
			//localstore.clearProfiles();
			localstore.stop();
		}catch(Exception e){
			e.printStackTrace();
		}
		ApiProxy.setEnvironmentForCurrentThread(null);
		ApiProxy.setDelegate(null);
	}
	
	private String csvFileNameToClassName(String fileName){
		File file = new File(fileName);
		String name = file.getName();
		String[] parts = name.split("\\.");
		return parts[0];
	}
	
	private void csvtest() throws Exception {
		
		String jdoClassName = "";
		
		startUP();
	
		PersistenceManager pm = PMF.get().getPersistenceManager();

		int line = 0;
		
		try{
			
			CSVReader reader = new CSVReader(new FileReader(csvFile));
			jdoClassName = csvFileNameToClassName(csvFile);
			Class<?> jdoClass = getClass(jdoClassName);
			
			System.out.println("csv file loaded.");
			line++;
			
			//csvファイルの1行目
			String [] headerLine;
			CSVtoJDOParser parser = new CSVtoJDOParser();
			if((headerLine = reader.readNext()) != null) {
				parser.parseHeaderLine(headerLine);
			}

			//csvファイルの2行目以降
			String [] nextLine;
			while ((nextLine = reader.readNext()) != null) {
				line++;
				System.out.println(line);
				
				HashMap<String, String> paramSet = parser.getParamsMap(nextLine);
				Object entityObj = setJDOParam(jdoClass, paramSet);

				if(parser.parentClassMethodName==null){
					//new
					pm.makePersistent(entityObj);
				}else{

					Key parentKey = getParentKey(parser.getParentPath(),nextLine);
					Object parentObj = pm.getObjectById(getClass(parser.parentClassName), parentKey);
					if(parentObj!=null){
						//update
						Method parentMethod = getMethodMap(parentObj).get(parser.parentClassMethodName);
						parentMethod.invoke(parentObj, entityObj);
					}

				}
			}
			System.out.println("jdo convert completed !");
			
		}catch(CSVtoJDOSetMethodException e){
			System.out.println("setter error ! " + csvFile + " line : "+ line);
			e.printStackTrace();
		}catch(CSVtoJDOClassConvertException e){
			System.out.println("class convert error ! " + csvFile + " line : "+ line);
			e.printStackTrace();
		}catch(FileNotFoundException e){
			System.out.println(csvFile + " not found.");
			e.printStackTrace();
		}catch(ClassNotFoundException e){
			System.out.println("JDO class \"" + jdoClassName + "\" not found.");
			e.printStackTrace();
		}catch(Exception e){
			System.out.println("error ! " + csvFile + " line : "+ line);
			e.printStackTrace();
		}
		
		pm.close();
		
		downEnd();
		
	}

	@SuppressWarnings("unused")
	private void newtest(){

		startUP();
		
		TestData e = new TestData();
		
		PersistenceManager pm = PMF.get().getPersistenceManager();
		try {
			pm.makePersistent(e);
			System.out.println(e);
		} finally {
			pm.close();
		}
		
		downEnd();
	}

	@SuppressWarnings("unused")
	private void deltest() throws Exception {
		
		startUP();
	
		PersistenceManager pm = PMF.get().getPersistenceManager();
		
		try{
			Query query = pm.newQuery(TestData.class);
			List<?> results = (List<?>) query.execute();
			if (results.iterator().hasNext()) {
				for (Object e : results) {
					pm.deletePersistent(e);
				}
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		
		pm.close();
		
		downEnd();
		
	}
}

class UsersException extends Exception {
	private static final long serialVersionUID = 1L;
	public UsersException(String str) {
		super(str);
	}
}

class CSVtoJDOClassConvertException extends UsersException {
	private static final long serialVersionUID = 1L;
	CSVtoJDOClassConvertException() {
		super("class convert error.");
	}
}

class CSVtoJDOSetMethodException extends UsersException {
	private static final long serialVersionUID = 1L;
	CSVtoJDOSetMethodException() {
		super("set method error.");
	}
}